'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _createClass = function () { function defineProperties(target, props) { for (var i = 0; i < props.length; i++) { var descriptor = props[i]; descriptor.enumerable = descriptor.enumerable || false; descriptor.configurable = true; if ("value" in descriptor) descriptor.writable = true; Object.defineProperty(target, descriptor.key, descriptor); } } return function (Constructor, protoProps, staticProps) { if (protoProps) defineProperties(Constructor.prototype, protoProps); if (staticProps) defineProperties(Constructor, staticProps); return Constructor; }; }();

var _clipperLib = require('clipper-lib');

var _clipperLib2 = _interopRequireDefault(_clipperLib);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function _toConsumableArray(arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) { arr2[i] = arr[i]; } return arr2; } else { return Array.from(arr); } }

function _classCallCheck(instance, Constructor) { if (!(instance instanceof Constructor)) { throw new TypeError("Cannot call a class as a function"); } }

_clipperLib2.default.Error = function (message) {
  console.warn('clipper error', message);
};

var CLIPPER = new _clipperLib2.default.Clipper();
var CLIPPER_OFFSET = new _clipperLib2.default.ClipperOffset();

var Shape = function () {
  function Shape() {
    var paths = arguments.length <= 0 || arguments[0] === undefined ? [] : arguments[0];
    var closed = arguments.length <= 1 || arguments[1] === undefined ? true : arguments[1];
    var capitalConversion = arguments.length <= 2 || arguments[2] === undefined ? false : arguments[2];
    var integerConversion = arguments.length <= 3 || arguments[3] === undefined ? false : arguments[3];

    _classCallCheck(this, Shape);

    this.paths = paths;
    if (capitalConversion) this.paths = this.paths.map(mapLowerToCapital);
    if (integerConversion) this.paths = this.paths.map(mapToRound);
    this.closed = closed;
  }

  _createClass(Shape, [{
    key: '_clip',
    value: function _clip(clipShape, type) {
      var solution = new _clipperLib2.default.PolyTree();

      CLIPPER.Clear();
      CLIPPER.AddPaths(this.paths, _clipperLib2.default.PolyType.ptSubject, this.closed);
      CLIPPER.AddPaths(clipShape.paths, _clipperLib2.default.PolyType.ptClip, clipShape.closed);
      CLIPPER.Execute(type, solution);

      var newShape = _clipperLib2.default.Clipper.PolyTreeToPaths(solution);
      return new Shape(newShape, this.closed);
    }
  }, {
    key: 'union',
    value: function union(clipShape) {
      return this._clip(clipShape, _clipperLib2.default.ClipType.ctUnion);
    }
  }, {
    key: 'difference',
    value: function difference(clipShape) {
      return this._clip(clipShape, _clipperLib2.default.ClipType.ctDifference);
    }
  }, {
    key: 'intersect',
    value: function intersect(clipShape) {
      return this._clip(clipShape, _clipperLib2.default.ClipType.ctIntersection);
    }
  }, {
    key: 'xor',
    value: function xor(clipShape) {
      return this._clip(clipShape, _clipperLib2.default.ClipType.ctXor);
    }
  }, {
    key: 'offset',
    value: function offset(_offset) {
      var options = arguments.length <= 1 || arguments[1] === undefined ? {} : arguments[1];
      var _options$jointType = options.jointType;
      var jointType = _options$jointType === undefined ? 'jtSquare' : _options$jointType;
      var _options$endType = options.endType;
      var endType = _options$endType === undefined ? 'etClosedPolygon' : _options$endType;
      var _options$miterLimit = options.miterLimit;
      var miterLimit = _options$miterLimit === undefined ? 2.0 : _options$miterLimit;
      var _options$roundPrecisi = options.roundPrecision;
      var roundPrecision = _options$roundPrecisi === undefined ? 0.25 : _options$roundPrecisi;


      CLIPPER_OFFSET.Clear();
      CLIPPER_OFFSET.ArcTolerance = roundPrecision;
      CLIPPER_OFFSET.MiterLimit = miterLimit;

      var offsetPaths = new _clipperLib2.default.Paths();
      CLIPPER_OFFSET.AddPaths(this.paths, _clipperLib2.default.JoinType[jointType], _clipperLib2.default.EndType[endType]);
      CLIPPER_OFFSET.Execute(offsetPaths, _offset);

      return new Shape(offsetPaths, true);
    }
  }, {
    key: 'scaleUp',
    value: function scaleUp(factor) {
      _clipperLib2.default.JS.ScaleUpPaths(this.paths, factor);

      return this;
    }
  }, {
    key: 'scaleDown',
    value: function scaleDown(factor) {
      _clipperLib2.default.JS.ScaleDownPaths(this.paths, factor);

      return this;
    }
  }, {
    key: 'lastPoint',
    value: function lastPoint() {
      if (this.paths.length === 0) {
        return;
      }

      var lastPath = this.paths[this.paths.length - 1];
      return this.closed ? lastPath[0] : lastPath[lastPath.length - 1];
    }
  }, {
    key: 'areas',
    value: function areas() {
      var areas = [];

      for (var i = 0; i < this.paths.length; i++) {
        var area = this.area(i);
        areas.push(area);
      }

      return areas;
    }
  }, {
    key: 'area',
    value: function area(index) {
      var path = this.paths[index];
      var area = _clipperLib2.default.Clipper.Area(path);
      return area;
    }
  }, {
    key: 'totalArea',
    value: function totalArea() {
      return this.areas().reduce(function (a, b) {
        return a + b;
      });
    }
  }, {
    key: 'reverse',
    value: function reverse() {
      _clipperLib2.default.Clipper.ReversePaths(this.paths);

      return this;
    }
  }, {
    key: 'tresholdArea',
    value: function tresholdArea(minArea) {
      // code not tested yet

      var _arr = [].concat(_toConsumableArray(this.paths));

      for (var _i = 0; _i < _arr.length; _i++) {
        var path = _arr[_i];
        var area = Math.abs(_clipperLib2.default.Clipper.Area(shape));

        if (area < minArea) {
          var index = this.paths.indexOf(path);
          this.splice(index, 1);
        }
      }
    }
  }, {
    key: 'join',
    value: function join(shape) {
      var _paths;

      (_paths = this.paths).push.apply(_paths, _toConsumableArray(shape.paths));

      return this;
    }
  }, {
    key: 'clone',
    value: function clone() {
      return new Shape(_clipperLib2.default.JS.Clone(this.paths), this.closed);
    }
  }, {
    key: 'shapeBounds',
    value: function shapeBounds() {
      var bounds = _clipperLib2.default.JS.BoundsOfPaths(this.paths);

      bounds.width = bounds.right - bounds.left;
      bounds.height = bounds.bottom - bounds.top;
      bounds.size = bounds.width * bounds.height;

      return bounds;
    }
  }, {
    key: 'pathBounds',
    value: function pathBounds(index) {
      var path = this.paths[index];

      var bounds = _clipperLib2.default.JS.BoundsOfPath(path);

      bounds.width = bounds.right - bounds.left;
      bounds.height = bounds.bottom - bounds.top;
      bounds.size = bounds.width * bounds.height;

      return bounds;
    }
  }, {
    key: 'clean',
    value: function clean(cleanDelta) {
      return new Shape(_clipperLib2.default.Clipper.CleanPolygons(this.paths, cleanDelta), this.closed);
    }
  }, {
    key: 'orientation',
    value: function orientation(index) {
      var path = this.paths[index];
      return _clipperLib2.default.Clipper.Orientation(path);
    }
  }, {
    key: 'pointInShape',
    value: function pointInShape(point) {
      var round = arguments.length <= 1 || arguments[1] === undefined ? false : arguments[1];

      if (round) point = roundVector(point);
      for (var i = 0; i < this.paths.length; i++) {
        var pointInPath = this.pointInPath(i, point);
        var orientation = this.orientation(i);

        if (!pointInPath && orientation || pointInPath && !orientation) {
          return false;
        }
      }

      return true;
    }
  }, {
    key: 'pointInPath',
    value: function pointInPath(index, point) {
      var round = arguments.length <= 2 || arguments[2] === undefined ? false : arguments[2];

      if (round) point = roundVector(point);
      var path = this.paths[index];
      var intPoint = { X: Math.round(point.X), Y: Math.round(point.Y) };

      return _clipperLib2.default.Clipper.PointInPolygon(intPoint, path) > 0;
    }
  }, {
    key: 'fixOrientation',
    value: function fixOrientation() {
      if (!this.closed) {
        return this;
      }

      if (this.totalArea() < 0) {
        this.reverse();
      }

      return this;
    }
  }, {
    key: 'removeOverlap',
    value: function removeOverlap() {
      if (this.closed) {
        var _shape = _clipperLib2.default.Clipper.SimplifyPolygons(this.paths, _clipperLib2.default.PolyFillType.pftNonZero);
        return new Shape(_shape, true);
      } else {
        return this;
      }
    }
  }, {
    key: 'seperateShapes',
    value: function seperateShapes() {
      var _this = this;

      var shapes = [];

      if (!this.closed) {
        var _iteratorNormalCompletion = true;
        var _didIteratorError = false;
        var _iteratorError = undefined;

        try {
          for (var _iterator = this.paths[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
            var path = _step.value;

            shapes.push(new Shape([path], false));
          }
        } catch (err) {
          _didIteratorError = true;
          _iteratorError = err;
        } finally {
          try {
            if (!_iteratorNormalCompletion && _iterator.return) {
              _iterator.return();
            }
          } finally {
            if (_didIteratorError) {
              throw _iteratorError;
            }
          }
        }
      } else {
        (function () {
          var map = new WeakMap();
          var outlines = [];
          var holes = [];

          for (var i = 0; i < _this.paths.length; i++) {
            var path = _this.paths[i];
            var orientation = _this.orientation(i);

            if (orientation) {
              var area = _this.area(i);
              map.set(path, { area: area, index: i });
              outlines.push(path);
            } else {
              holes.push(path);
            }
          }

          outlines.sort(function (a, b) {
            return map.get(a).area > map.get(b).area;
          });

          var _iteratorNormalCompletion2 = true;
          var _didIteratorError2 = false;
          var _iteratorError2 = undefined;

          try {
            for (var _iterator2 = outlines[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
              var outline = _step2.value;

              var _shape2 = [outline];

              var _map$get = map.get(outline);

              var index = _map$get.index;

              var _arr2 = [].concat(holes);

              for (var _i2 = 0; _i2 < _arr2.length; _i2++) {
                var hole = _arr2[_i2];
                var pointInHole = _this.pointInPath(index, hole[0]);
                if (pointInHole) {
                  _shape2.push(hole);

                  var _index = holes.indexOf(hole);
                  holes.splice(_index, 1);
                }
              }

              shapes.push(new Shape(_shape2, true));
            }
          } catch (err) {
            _didIteratorError2 = true;
            _iteratorError2 = err;
          } finally {
            try {
              if (!_iteratorNormalCompletion2 && _iterator2.return) {
                _iterator2.return();
              }
            } finally {
              if (_didIteratorError2) {
                throw _iteratorError2;
              }
            }
          }
        })();
      }

      return shapes;
    }
  }, {
    key: 'mapToLower',
    value: function mapToLower() {
      return this.paths.map(mapCapitalToLower);
    }
  }]);

  return Shape;
}();

exports.default = Shape;


function mapCapitalToLower(path) {
  return path.map(vectorToLower);
}

function vectorToLower(_ref) {
  var X = _ref.X;
  var Y = _ref.Y;

  return { x: X, y: Y };
}

function mapLowerToCapital(path) {
  return path.map(vectorToCapital);
}

function vectorToCapital(_ref2) {
  var x = _ref2.x;
  var y = _ref2.y;

  return { X: x, Y: y };
}

function mapToRound(path) {
  return path.map(roundVector);
}

function roundVector(_ref3) {
  var X = _ref3.X;
  var Y = _ref3.Y;

  return { X: Math.round(X), Y: Math.round(Y) };
}
module.exports = exports['default'];